import matplotlib.pyplot as plt
import os
import subprocess
import pandas as pd
import numpy as np
import io
cwd = os.getcwd()


# function that runs a command line command and returns the output
# it also takes a list of arguments to pass to the command
def run_command(command, args):
    # convert command and args to a string
    try:
        cmd = [command] + args
        print("Running command:", ' '.join(cmd))
        result = subprocess.run(cmd, check=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE, text=True)
        return result.stdout, result.stderr
    except subprocess.CalledProcessError as e:
        print('stdout output:\n', e.stdout)
        print('stderr output:\n', e.stderr)
        print("Error executing command:", e)


# function that builds a string from a dictionary of arguments

def build_args(args):
    args_list = []
    for key, value in args.items():
        args_list.append(key)
        args_list.append(value)
    return args_list


# function

# example command to run:
# nsys profile -o cuperftest_profile ./cuperftest --prec f --n_runs 10 --method 1 --N1 256 --N2 256 --N3 256 --M 1E8 --tol 1E-6
# example arguments
args = {"--prec": "f",
        "--n_runs": "5",
        "--method": "0",
        "--sort": "1",
        "--N1": "16777216",
        # "--N1": "256",
        # "--N2": "256",
        # "--N3": "256",
        "--kerevalmethod": "1",
        "--M": "1E8",
        "--tol": "1E-6"}
# iterate over tol from 1E-6 to 1E-1

warmup = {"--prec": "f",
        "--n_runs": "1",
        "--method": "0",
        "--N1": "256",
        # "--N2": "256",
        # "--N3": "256",
        "--M": "256",
        "--tol": "1E-1"}
cmd = ["profile", "--force-overwrite", "true", "-o", "cuperftest_profile", cwd + "/cuperftest"] + build_args(warmup)
print("Warmup")
stdout, stderr = run_command("nsys", cmd)
print("Benchmarking")
if stderr != '':
    print(stderr)
    exit(0)
for precision in ['d']:
    print(f"precision: {precision}")
    for dim in range(1, 2):
        if dim == 1:
            args["--N1"] = "16777216"
        if dim == 2:
            args["--N1"] = "256"
            args["--N2"] = "256"
        if dim == 3:
            args["--N1"] = "256"
            args["--N2"] = "256"
            args["--N3"] = "256"
        args["--prec"] = precision
        max_range = 16 if args["--prec"] == "d" else 7
        if precision == 'd' and dim == 3:
            max_range = 6
        print(f"dimensions {dim}")
        data = {
            'method': [],
            'throughput': [],
            'tolerance': [],
            # 'setpts': [],
            'exec': [],
        }
        for i in range(1, max_range):
            args["--tol"] = "1E-" + ("0" if i < 10 else "") + str(i)
            print("Running with tol = 1E-" + str(i))
            for method in ['2', '1']:
                args["--method"] = method
                if method == '0':
                    data['method'].append('auto')
                elif method == '1':
                    data['method'].append('GM')
                elif method == '2':
                    data['method'].append('SM')
                elif method == '4':
                    data['method'].append('BLOCK')
                print("Method " + data['method'][-1])
                cmd = ["profile", "--force-overwrite", "true", "-o", "cuperftest_profile", cwd + "/cuperftest"] + build_args(args)
                stdout, stderr = run_command("nsys", cmd)
                if stderr != '':
                    print(stderr)
                    exit(0)
                # skip all lines starting with # in stdout
                conf = [x for x in stdout.splitlines() if x.startswith("#")]
                print('\n'.join(conf))
                stdout = [x for x in stdout.splitlines() if not x.startswith("#")][:7]
                if stdout[0].startswith("bin"):
                    print(stdout[0])
                    stdout = stdout[1:]

                stdout = '\n'.join(stdout)
                # convert stdout to a dataframe from csv string
                dt = pd.read_csv(io.StringIO(stdout), sep=',')
                setpts = dt[dt["event"].str.contains("setpts")]['nupts/s'].sum() # it is only one row it extracts the value
                exec = dt[dt["event"].str.contains("exec")]['nupts/s'].sum() # it is only one row it extracts the value
                # print(f'setpts pts/s: {setpts}')
                # print(f'exec pts/s: {exec}')
                cmd = ["stats", "--force-overwrite=true", "--force-export=true", "--report", "cuda_gpu_trace", "--report", "cuda_gpu_kern_sum", "cuperftest_profile.nsys-rep",
                       "--format=csv", "--output", "cuperftest"]
                stdout, _ = run_command("nsys", cmd)
                # remove format from cmd
                cmd = cmd[:-3]
                # print(run_command("nsys", cmd))
                # print(csv)
                dt = pd.read_csv("./cuperftest_cuda_gpu_trace.csv")
                # print(dt)
                # sum the "Total Time" column of the ones that contain "fft" in name
                # print(dt[dt["Name"].str.contains("fft") & ~dt["Name"].str.contains("cufinufft")])
                total_fft = dt[dt["Name"].str.contains("fft") & ~dt["Name"].str.contains("cufinufft")]['Duration (ns)'].sum()
                # print(f'total_fft: {total_fft}')
                # drop all the rows with spread not in "Name"
                dt = dt[dt["Name"].str.contains("cufinufft::spreadinterp::spread")]
                # print(dt)
                # exit(0)
                # sort dt by column "Time (%)"
                total_spread = dt['Duration (ns)'].sum() - total_fft
                # print(f'total_spread: {total_spread}')
                if total_fft > total_spread:
                    print("Warning: total_fft > total_spread")
                    # exit(0)
                # pt/s
                throughput = float(args['--M']) * float(args['--n_runs']) * 1_000_000_000 / total_spread
                print(f'throughput: {throughput}')
                data['throughput'].append(throughput)
                data['tolerance'].append(args['--tol'])
                # data['setpts'].append(setpts)
                data['exec'].append(exec)
        df = pd.DataFrame(data)
        # Pivot the DataFrame
        pivot_df = df.pivot(index='tolerance', columns='method')
        # print(pivot_df)
        # scale the throughput SM by GM
        # pivot_df['throughput', 'SM'] /= pivot_df['throughput', 'GM']
        # pivot_df['throughput', 'GM'] /= pivot_df['throughput', 'GM']
        # scale setpts SM by GM
        # pivot_df['exec', 'SM'] /= pivot_df['exec', 'GM']
        # pivot_df['exec', 'GM'] /= pivot_df['exec', 'GM']
        # remove the GM column
        # pivot_df.drop(('throughput', 'GM'), axis=1, inplace=True)
        pivot_df.drop(('exec', 'GM'), axis=1, inplace=True)
        pivot_df.drop(('exec', 'SM'), axis=1, inplace=True)
        print(pivot_df)
exit(0)
# Plot
pivot_df.plot(kind='bar', figsize=(10, 7))
# Find the minimum throughput value
min_val = min(pivot_df[('throughput', 'SM')].min(), pivot_df[('throughput', 'GM')].min())
max_val = max(pivot_df[('throughput', 'SM')].max(), pivot_df[('throughput', 'GM')].max())
print(min_val, max_val)
plt.ylim(min_val * .90, max_val * 1.1)
# plt.ylim(.8, 1.2)

# Calculate the smallest power of 10
# min_pow_10 = 10 ** np.floor(np.log10(min_throughput))

# Adjust the plot's y-axis limits
# plt.ylim(df['throughput'].min()*.99, df['throughput'].max() * 1.009)  # Adding 10% for upper margin

# plot an horizontal line at 1 with label "GM"
# plt.axhline(y=1, color='k', linestyle='--', label='GM')
plt.xlabel('Tolerance')
plt.ylabel('Throughput')
plt.title('Throughput by Tolerance and Method')
plt.legend(title='Method')
plt.tight_layout()
plt.show()
plt.xlabel("Tolerance")
plt.ylabel("Points/s")
plt.savefig("bench.png")
plt.savefig("bench.svg")
plt.savefig("bench.pdf")
plt.show()



import pytest

import numpy as np
import finufft

import utils


SHAPES = [(7,), (8,), (7, 7), (7, 8), (8, 8), (7, 7, 7), (7, 8, 8), (8, 8, 8)]
N_PTS = [10, 11]
N_TRANS = [(), (2,)]
DTYPES = [np.complex64, np.complex128]
OUTPUT_ARGS = [False, True]


@pytest.mark.parametrize("dtype", DTYPES)
@pytest.mark.parametrize("shape", SHAPES)
@pytest.mark.parametrize("n_pts", N_PTS)
@pytest.mark.parametrize("n_trans", N_TRANS)
@pytest.mark.parametrize("output_arg", OUTPUT_ARGS)
def test_finufft1_simple(dtype, shape, n_pts, n_trans, output_arg):
    dim = len(shape)

    funs = {1: finufft.nufft1d1,
            2: finufft.nufft2d1,
            3: finufft.nufft3d1}

    fun = funs[dim]

    pts, coefs = utils.type1_problem(dtype, shape, n_pts, n_trans)

    # See if it can handle square sizes from ints
    if all(n == shape[0] for n in shape):
        _shape = shape[0]
    else:
        _shape = shape

    if not output_arg:
        sig = fun(*pts, coefs, _shape)
    else:
        sig = np.empty(n_trans + shape, dtype=dtype)
        fun(*pts, coefs, out=sig)

    utils.verify_type1(pts, coefs, shape, sig, 1e-6)

@pytest.mark.parametrize("dtype", DTYPES)
@pytest.mark.parametrize("shape", SHAPES)
@pytest.mark.parametrize("n_pts", N_PTS)
@pytest.mark.parametrize("n_trans", N_TRANS)
@pytest.mark.parametrize("output_arg", OUTPUT_ARGS)
def test_finufft2_simple(dtype, shape, n_pts, n_trans, output_arg):
    dim = len(shape)

    funs = {1: finufft.nufft1d2,
            2: finufft.nufft2d2,
            3: finufft.nufft3d2}

    fun = funs[dim]

    pts, sig = utils.type2_problem(dtype, shape, n_pts, n_trans)

    if not output_arg:
        coefs = fun(*pts, sig)
    else:
        coefs = np.empty(n_trans + (n_pts,), dtype=dtype)
        fun(*pts, sig, out=coefs)

    utils.verify_type2(pts, sig, coefs, 1e-6)

@pytest.mark.parametrize("dtype", DTYPES)
@pytest.mark.parametrize("dim", list(set(len(shape) for shape in SHAPES)))
@pytest.mark.parametrize("n_source_pts", N_PTS)
@pytest.mark.parametrize("n_target_pts", N_PTS)
@pytest.mark.parametrize("n_trans", N_TRANS)
@pytest.mark.parametrize("output_arg", OUTPUT_ARGS)
def test_finufft3_simple(dtype, dim, n_source_pts, n_target_pts, n_trans, output_arg):
    funs = {1: finufft.nufft1d3,
            2: finufft.nufft2d3,
            3: finufft.nufft3d3}

    fun = funs[dim]

    source_pts, source_coefs, target_pts = utils.type3_problem(dtype,
            dim, n_source_pts, n_target_pts, n_trans)

    if not output_arg:
        target_coefs = fun(*source_pts, source_coefs, *target_pts)
    else:
        target_coefs = np.empty(n_trans + (n_target_pts,), dtype=dtype)
        fun(*source_pts, source_coefs, *target_pts, out=target_coefs)

    utils.verify_type3(source_pts, source_coefs, target_pts, target_coefs, 1e-6)

def test_finufft_simple_errors():
    with pytest.raises(RuntimeError, match="x dtype should be"):
        finufft.nufft1d1(np.zeros(1, "int64"), np.zeros(1), (4,))

    with pytest.raises(RuntimeError, match="n_modes must be"):
        finufft.nufft1d1(np.zeros(1), np.zeros(1), "")

    with pytest.raises(RuntimeError, match="n_modes dimension does not match"):
        finufft.nufft1d1(np.zeros(1), np.zeros(1), (2, 2))

    with pytest.raises(RuntimeError, match="elements of input n_modes must"):
        finufft.nufft1d1(np.zeros(1), np.zeros(1), (2.5,))

    with pytest.raises(RuntimeError, match="type 1 input must supply n_modes"):
        finufft.nufft1d1(np.zeros(1), np.zeros(1))

    with pytest.raises(RuntimeError, match="c.size must be divisible by x.size"):
        finufft.nufft1d1(np.zeros(2), np.zeros(3, np.complex128), 4)

    with pytest.raises(RuntimeError, match="type 2 input dimension must be either dim"):
        finufft.nufft1d2(np.zeros(1), np.zeros((2, 2, 2), np.complex128))

    with pytest.raises(RuntimeError, match="type 2 input dimension must be either dim"):
        finufft.nufft1d1(np.zeros(1), np.zeros(1, np.complex128), 4, out=np.zeros((2, 2, 4), np.complex128))

    with pytest.raises(RuntimeError, match="input n_trans and output n_trans do not match"):
        finufft.nufft1d1(np.zeros(1), np.zeros((3, 1), np.complex128), 4, out=np.zeros((2, 4), np.complex128))

    with pytest.raises(RuntimeError, match="input n_modes and output n_modes do not match"):
        finufft.nufft1d1(np.zeros(1), np.zeros(1, np.complex128), 4, out=np.zeros((3), np.complex128))



import commands
import re
import sys

# Test set
tests = []

#tests.append({"dim":1, "M":1e8, "N":1e8, "tol":1e-3})
#tests.append({"dim":1, "M":1e8, "N":1e8, "tol":1e-8})
#tests.append({"dim":1, "M":1e8, "N":1e8, "tol":1e-15})

#tests.append({"dim":3, "M":1e7, "N":1e7, "tol":1e-3})
#tests.append({"dim":3, "M":1e7, "N":1e7, "tol":1e-8})

tests.append({"dim":3, "M":1e5, "N":1e5, "tol":1e-15})
tests.append({"dim":3, "M":1e6, "N":1e6, "tol":1e-15})
tests.append({"dim":3, "M":1e7, "N":1e7, "tol":1e-15})


# Flags to send
flags = "1"

# Make flags (eg OMP=OFF)
makeflags = ""

# Command template
cmdtemplate = "test/spreadtestnd %(dim)d %(M)g %(N)g %(tol)g " + flags

# Regexps to use
spreadre = re.compile(r'dir=1.*\n.*\s(?P<speed>\d+[\d\.e+-]*) pts/s.*\n.*rel err.*\s(?P<err>\d+[\d\.e+-]*)')
interpre = re.compile(r'dir=2.*\n.*\s(?P<speed>\d+[\d\.e+-]*) pts/s.*\n.*rel err.*\s(?P<err>\d+[\d\.e+-]*)')

commit_A = sys.argv[1]
commit_B = sys.argv[2]

print "* Comparing commits '%s' and '%s'" % (commit_A, commit_B)

# Find out where we are, so we can come back
gitout = commands.getoutput("git status -b --porcelain")
pos = re.match("## (.+)\n", gitout).group(1)
if re.match("HEAD ", pos):
    home = commands.getoutput("git rev-parse HEAD")
    print "* Seems we are currently in detached head, will return to commit %s" % home
else:
    home = re.match("([^ \.]+)", pos).group(1)
    print "* Will return to branch %s" % home

# Command runner
def runCommand(cmd):
    print "> " + cmd
    status, output = commands.getstatusoutput(cmd)
    print output
    return output
    
# Test runner
def runTests():
    print "Running tests..."
    results = []
    for params in tests:
        cmd = cmdtemplate % params
        # Best of 3
        interp_speed = 0
        interp_err = 0
        spread_speed = 0
        spread_err = 0
        for i in [1,2,3]:
            output = runCommand(cmd).rstrip()
            ms = spreadre.search(output)
            mi = interpre.search(output)
            interp_speed = max(interp_speed, mi.group("speed"))
            interp_err = max(interp_err, mi.group("err"))
            spread_speed = max(spread_speed, ms.group("speed"))
            spread_err = max(spread_err, ms.group("err"))            
        results.append({"cmd":cmd,
                        "interp_speed":interp_speed,
                        "interp_err":interp_err,
                        "spread_speed":spread_speed,
                        "spread_err":spread_err })
    return results

# Code checkout machinery
def checkoutandmake(commit):
    if commit == "local":
        # Do nothin, just make it
        pass
    else:
        # Stash current source tree and check out commit
        print "Stashing changes and checking out %s..." % commit
        runCommand("git stash save")
        runCommand("git checkout %s" % commit)
    print "Making..."
    print commands.getoutput("make clean test/spreadtestnd " + makeflags)
    
def restore(commit, home):
    if commit == "local":
        # We just tested local, so do nothing
        pass
    else:
        # Return home and pop stash
        print "Checking out %s and popping stash..." % home
        runCommand("git checkout %s" % home)        
        runCommand("git stash pop")
    
# Run tests    
print "* Testing %s" % commit_A    
checkoutandmake(commit_A)
res_A = runTests()
restore(commit_A, home)

print "* Testing %s" % commit_B    
checkoutandmake(commit_B)
res_B = runTests()
restore(commit_B, home)

# Present results

format = "%-15s | %-15s | %-15s | %-15s | %-15s "
header = format % ("Commit", "Spread spd", "Interp spd", "Spread err", "Interp err")
print ""
print "Make flags: " + makeflags
print ""

for idx,params in enumerate(tests):
    print "=== Test: dim=%(dim)d, M=%(M)g, N=%(N)g, tol=%(tol)g" % params
    print header
    c = commit_A
    r = res_A[idx]
    print format % (c, r["spread_speed"]+" pts/s", r["interp_speed"]+" pts/s", r["spread_err"], r["interp_err"])
    c = commit_B
    r = res_B[idx]
    print format % (c, r["spread_speed"]+" pts/s", r["interp_speed"]+" pts/s", r["spread_err"], r["interp_err"])

    spread_speedup = float(res_B[idx]["spread_speed"]) / float(res_A[idx]["spread_speed"])*100 - 100
    spread_sign = "+" if spread_speedup > 0 else ""
    spread_speedup = spread_sign + "%.1f" % spread_speedup + "%"    
    interp_speedup = float(res_B[idx]["interp_speed"]) / float(res_A[idx]["interp_speed"])*100 - 100   
    interp_sign = "+" if interp_speedup > 0 else ""
    interp_speedup = interp_sign + "%.1f" % interp_speedup + "%"    
    print format % ("", spread_speedup, interp_speedup, "", "")

    print ""



import pytest

import numpy as np

from cufinufft import Plan, _compat

import utils


def test_type1_ordering(to_gpu, to_cpu, dtype=np.float32, shape=(16, 16, 16), M=4096, tol=1e-3):
    complex_dtype = utils._complex_dtype(dtype)

    k, c = utils.type1_problem(dtype, shape, M)

    k_gpu = to_gpu(k)
    c_gpu = to_gpu(c)

    plan = Plan(1, shape, eps=tol, dtype=complex_dtype)

    plan.setpts(*k_gpu)

    out = np.empty(shape, dtype=complex_dtype, order="F")

    out_gpu = to_gpu(out)

    with pytest.raises(TypeError, match="following requirement: C") as err:
        plan.execute(c_gpu, out=out_gpu)



fast = 'new.txt'
slow = 'old.txt'


def read_data(filename):
    data = [0] * 17
    with open(filename) as f1:
        nspread = 0
        speed = 0
        for line in f1:
            if 'nspread' in line:
                nspread = int(line.split('=')[-1])
            if 'pts/s' in line:
                speed = float(line.split(' ')[12])
            data[nspread] = speed
    return data

# compute relative increment in percentage between two numbers


vec = read_data(fast)[2:]
old = read_data(slow)[2:]

# 1 : slow = x : fast
# x = (1 - slow/fast) * 100
i = 2
for vec, old in zip(vec, old):
    diff = (1 - old/vec)*100
    print(f'nspread={i:02d} delta={diff:.3f}%')
    i+=1



# convert DFM's simple demo to JFM interface, include modeord test.
# Barnett 10/25/17. Adde upsampfac, 6/18/18

import time
import finufft
import numpy as np

# print finufft.nufft1d1.__doc__

np.random.seed(42)

acc = 1.e-9
iflag = 1
N = int(1e6)
M = int(1e5)
x = np.random.uniform(-np.pi, np.pi, M)
c = np.random.randn(M) + 1.j * np.random.randn(M)
F = np.zeros([N], dtype=np.complex128)       # allocate F (modes out)

strt = time.time()
F = finufft.nufft1d1(x, c, N, eps=acc, isign=iflag, debug=1, spread_debug=1)
print("Finished nufft in {0:.2g} seconds. Checking..."
      .format(time.time()-strt))

n = 142519      # mode to check
Ftest = 0.0
# this is so slow...
for j in range(M):
    Ftest += c[j] * np.exp(n * x[j] * 1.j)
Fmax = np.max(np.abs(F))
err = np.abs((F[n + N // 2] - Ftest) / Fmax)
print("Error relative to max of F: {0:.2e}".format(err))

# now test FFT mode output version, overwriting F...
strt = time.time()
finufft.nufft1d1(x, c, out=F, eps=acc, isign=iflag, modeord=1)
print("Finished nufft in {0:.2g} seconds (modeord=1)"
      .format(time.time()-strt))
err = np.abs((F[n] - Ftest) / Fmax)   # now zero offset in F array
print("Error relative to max of F: {0:.2e}".format(err))

# now test low-upsampfac (sigma) version...
strt = time.time()
Ftest2 = finufft.nufft1d1(x, c, N, F, acc, iflag, upsampfac=1.25)
print(Ftest2 is F)
print("Finished nufft in {0:.2g} seconds (upsampfac=1.25)"
      .format(time.time()-strt))
err = np.abs((Ftest2[n + N // 2] - Ftest) / Fmax)   # now zero offset in F array
print("Error relative to max of F: {0:.2e}".format(err))



import json

matrix = {
    "include": []
}

python_versions = ["3.8", "3.11"]

combinations = {
    "ubuntu-22.04": {
        "compiler": ["llvm", "gcc"],
        "arch_flags": ["-march=native", "-march=x86-64"]
    },
    "windows-2022": {
        "compiler": ["msvc", "llvm"],
        "arch_flags": ["/arch:AVX2", "/arch:SSE2"]
    },
    "macos-13": {
        "compiler": ["llvm", "gcc-14"],
        "arch_flags": ["-march=native", "-march=x86-64"]
    }
}

for platform in combinations.keys():
    for python_version in python_versions:
        for compiler in combinations[platform]["compiler"]:
            for arch_flag in combinations[platform]["arch_flags"]:
                matrix["include"].append({
                    "os": platform,
                    "python-version": python_version,
                    "compiler": compiler,
                    "arch_flags": arch_flag
                })

json_str = json.dumps(matrix, ensure_ascii=False)
print(json_str)
